##################################
##### Tests for otter notebook #####
##################################

import unittest
import sys
import os
import shutil
import subprocess
import json
import requests
import nbformat

from glob import glob
from subprocess import PIPE
from unittest.mock import patch
from unittest import mock

from otter import Notebook
from otter.check.logs import LogEntry, EventType, Log
from otter.check.notebook import _OTTER_LOG_FILENAME
from otter.check import notebook

from . import TestCase

TEST_FILES_PATH = "test/test-notebook/"


# functions used in one of the tests below
def square(x):
    return x ** 2

def negate(x):
    return not x

def mocked_requests_get(*args, **kwargs):
    class MockResponse:
        def __init__(self, json_data, status_code):
            self.text = json_data
            self.status_code = status_code

        def json(self):
            return self.json_data

    if args[0] == "http://some.url/auth/google":
        return MockResponse({"text": "value1"}, 200)
    elif args[0] == "http://some.url/submit":
        return MockResponse({"text": "this is the fake post response"}, 200)

    return MockResponse(None, 404)

def mock_auth_get():
    class AuthResponse:
        def __init__(self, json_data, status_code):
            self.content = json_data
            self.status_code = status_code

        def content(self):
            return self

        def decode(self, type):
            return "fakekey"

    str = "fakekey"
    content = str.encode("utf-8", "strict")
    response = AuthResponse(content, 2)

    return response


class TestNotebook(TestCase):
    """
    Test cases for the ``Notebook`` class
    """
    @mock.patch('builtins.input', return_value='fakekey')
    def test_check(self, mock_input):
        """
        Checks that the otter.Notebook class works correctly
        """
        grader = Notebook(tests_dir=TEST_FILES_PATH + "tests")

        def square(x):
            return x ** 2

        def negate(x):
            return not x

        global_env = {
            "square": square,
            "negate": negate
        }

        for q_path in glob(TEST_FILES_PATH + "tests/*.py"):
            q = os.path.split(q_path)[1][:-3]
            result = grader.check(q, global_env=global_env)
            if q != "q2":
                self.assertEqual(result.grade, 1, "Test {} failed".format(q))
            else:
                self.assertEqual(result.grade, 0, "Test {} passed".format(q))

    def test_check_no_env(self):
        """
        Checks that the ``Notebook`` class works correctly, with no global env input
        """
        grader = Notebook(tests_dir=TEST_FILES_PATH + "tests")

        for q_path in glob(TEST_FILES_PATH + "tests/*.py"):
            q = os.path.split(q_path)[1][:-3]
            result = grader.check(q)  # global_env=global_env)
            if q != "q2":
                self.assertEqual(result.grade, 1, f"Test {q} expected to pass but failed:\n{result}")
            else:
                self.assertEqual(result.grade, 0, f"Test {q} expected to fail but passed:\n{result}")

    def test_check_raise_exception(self):
        """
        Checks that the otter.Notebook class check method correctly raises Exceptions
        when things go wrong
        """
        grader = Notebook(tests_dir=TEST_FILES_PATH + "tests")
        global_env = 0
        for q_path in glob(TEST_FILES_PATH + "tests/*.py"):
            q = os.path.split(q_path)[1][:-3]
            self.assertRaises(AttributeError,
                              lambda: grader.check(q, global_env=global_env))

    def test_check_all_repr(self):
        """
        Checks that the representation of results as strings is correct
        """
        grader = Notebook(tests_dir=TEST_FILES_PATH + "tests")

        output = str(grader.check_all())
        output2 = grader.check_all()

        # checks each question substring
        output_lst = [
            'q1 results: All test cases passed!\n',
            'q2 results:\n    q2 - 1 result:\n        Trying:',
            'q3 results: All test cases passed!\n',
            'q4 results: All test cases passed!\n',
            'q5 results: All test cases passed!'
        ]

        for result in output_lst:
            self.assertTrue(output.count(result) == 1, f"Expected output to contain '{result}':\n{output}")

    def test_to_pdf_with_nb_path(self):
        """
        Checks for existence of notebook PDF
        This test is the general use case WITH a specified notebook path
        """
        nb = nbformat.v4.new_notebook()
        text = """\
        This is an auto-generated notebook."""
        nb['cells'] = [nbformat.v4.new_markdown_cell(text)]
        with open(TEST_FILES_PATH + 'test-nb.ipynb', "w") as f:
            nbformat.write(nb, f)
        grader = Notebook(tests_dir=TEST_FILES_PATH + "tests")
        grader.to_pdf(TEST_FILES_PATH + "test-nb.ipynb", filtering=False)

        self.assertTrue(os.path.isfile(TEST_FILES_PATH + "test-nb.pdf"))
        # cleanup
        os.remove(TEST_FILES_PATH + 'test-nb.ipynb')
        os.remove(TEST_FILES_PATH + "test-nb.pdf")

    def test_to_pdf_without_nb_path_case2_fail(self):
        """
        Checks for correct error scenario for to_pdf method
        This test is for when np_path is set to None and multiple
        IPYNB notebooks exist in the working directory.
        """
        nb1 = nbformat.v4.new_notebook()
        nb2 = nbformat.v4.new_notebook()
        text = """\
                This is an auto-generated notebook."""
        nb1['cells'] = [nbformat.v4.new_markdown_cell(text)]
        nb2['cells'] = [nbformat.v4.new_markdown_cell(text)]
        with open('test-nb1.ipynb', "w") as f:
            nbformat.write(nb1, f)
        with open('test-nb2.ipynb', "w") as f:
            nbformat.write(nb2, f)
        grader = Notebook(tests_dir=TEST_FILES_PATH + "tests")
        self.assertRaises(AssertionError,
                          lambda: grader.to_pdf(nb_path=None, filtering=False))
        os.remove('test-nb1.ipynb')
        os.remove('test-nb2.ipynb')

    def test_export_pass(self):
        """
        Checks export contents for existence of PDF and equality of zip
        """
        nb = nbformat.v4.new_notebook()
        text = """\
                        This is an auto-generated notebook."""
        nb['cells'] = [nbformat.v4.new_markdown_cell(text)]
        with open(TEST_FILES_PATH + 'test-nb.ipynb', "w") as f:
            nbformat.write(nb, f)
        correct_directory = TEST_FILES_PATH + 'export-correct/'
        os.mkdir(correct_directory)
        with open(correct_directory + 'test-nb.ipynb', "w") as f:
            nbformat.write(nb, f)
        grader = Notebook(tests_dir=TEST_FILES_PATH + "tests")
        grader.export(TEST_FILES_PATH + "test-nb.ipynb", filtering=False)

        self.assertTrue(os.path.isfile(TEST_FILES_PATH + "test-nb.pdf"))
        zips = glob(TEST_FILES_PATH + "test-nb*.zip")
        assert len(zips) == 1
        with self.unzip_to_temp(zips[0]) as unzipped_dir:
            # breakpoint()
            os.remove(unzipped_dir + '/test/test-notebook/test-nb.pdf')
            self.assertDirsEqual(
                unzipped_dir + '/test/test-notebook/',
                TEST_FILES_PATH + "export-correct",
                ignore_ext=[".pdf"]
            )

        # cleanup
        os.remove(correct_directory + "test-nb.ipynb")
        os.rmdir(correct_directory)
        os.remove(TEST_FILES_PATH + "test-nb.ipynb")
        os.remove(TEST_FILES_PATH + "test-nb.pdf")
        os.remove(zips[0])

    def test_export_without_nb_path_case2_fail(self):
        """
        Checks for correct error scenario for export method
        This test is for when nb_path is set to None and
        there are multiple IPYNB notebooks in working directory
        """
        nb1 = nbformat.v4.new_notebook()
        nb2 = nbformat.v4.new_notebook()
        text = """\
                        This is an auto-generated notebook."""
        nb1['cells'] = [nbformat.v4.new_markdown_cell(text)]
        nb2['cells'] = [nbformat.v4.new_markdown_cell(text)]
        with open('test-nb1.ipynb', "w") as f:
            nbformat.write(nb1, f)
        with open('test-nb2.ipynb', "w") as f:
            nbformat.write(nb2, f)
        grader = Notebook(tests_dir=TEST_FILES_PATH + "tests")
        self.assertRaises(AssertionError,
                          lambda: grader.export(nb_path=None, filtering=False))
        os.remove('test-nb1.ipynb')
        os.remove('test-nb2.ipynb')

    def test_export_without_nb_path_case3(self):
        """
        Checks for correct error scenario for export method
        This test is for when nb_path is set to None and
        there are no IPYNB notebooks in the working directory
        """
        files_in_directory = os.listdir('./')
        notebooks = [file for file in files_in_directory if file.endswith(".ipynb")]
        for file in notebooks:
            os.remove('./' + file)
        grader = Notebook()
        self.assertRaises(ValueError,
                          lambda: grader.export(nb_path=None, filtering=False))

    def test_export_multiple_otter_error(self):
        """
        Checks export for error scenario for export method
        This test should pass when export successfully raises an
        AssertionError for the case when the directory contains
        multiple .otter files.
        """
        grader = Notebook(tests_dir=TEST_FILES_PATH + "tests")
        self.assertRaises(ValueError, lambda: grader.export(nb_path=None, filtering=False))

    @patch.object(LogEntry, "shelve")
    def test_nb_log(self, mock_log):
        """
        Checks existence of log when running nb
        """

        mock_log.return_value = LogEntry(EventType.CHECK)
        grader = Notebook(tests_dir=TEST_FILES_PATH + "tests")
        output = grader.check_all()

        self.assertTrue(os.path.isfile(_OTTER_LOG_FILENAME))

    def test_colab(self):
        """
        Checks that the ``Notebook`` class correctly disables methods on Google Colab.
        """
        tests_dir = TEST_FILES_PATH + "tests"
        with mock.patch("otter.check.notebook.os.path.isdir") as mocked_isdir:
            mocked_isdir.return_value = False
            with self.assertRaises(ValueError, msg=f"Tests directory {tests_dir} does not exist"):
                grader = Notebook(tests_dir=tests_dir, colab=True)

            mocked_isdir.assert_called_once_with(tests_dir)

            mocked_isdir.return_value = True
            grader = Notebook(tests_dir=tests_dir, colab=True)

            # check for appropriate errors
            with self.assertRaises(RuntimeError, msg="This method is not compatible with Google Colab"):
                grader.run_plugin()

            with self.assertRaises(RuntimeError, msg="This method is not compatible with Google Colab"):
                grader.to_pdf()

            with self.assertRaises(RuntimeError, msg="This method is not compatible with Google Colab"):
                grader.add_plugin_files()

            with self.assertRaises(RuntimeError, msg="This method is not compatible with Google Colab"):
                grader.export()

    def tearDown(self):
        for i in range(1, 7):
            file = "demofile{}.otter".format(i)
            if os.path.isfile(file):
                os.remove(file)
        if os.path.isfile(_OTTER_LOG_FILENAME):
            os.remove(_OTTER_LOG_FILENAME)
        if os.path.exists("hw00.pdf"):
            os.remove("hw00.pdf")
        if os.path.exists("hw00.zip"):
            os.remove("hw00.zip")



import nbformat
import re

from collections import namedtuple

from .constants import BEGIN_REGEX, END_REGEX

from ..constants import BLOCK_QUOTE
from ..utils import get_source, get_spec


CELL_CREATORS = {
    "markdown": nbformat.v4.new_markdown_cell,
    "code": nbformat.v4.new_code_cell,
}


def create_cell(cell_type, source):
    """
    """
    return CELL_CREATORS[cell_type](source)


def rmd_to_cells(rmd_string):
    """
    Converts an Rmd document as a string into a list of ``Cell`` objects for easier handling with
    code designed originally for Jupyter notebooks.

    Args:
        rmd_string (``str``): the Rmd document as a string

    Returns:
        ``list`` of ``Cell``: the "cells" represented named tuples
    """
    cells, cell_lines, cell_type, in_block, in_begin = [], [], "markdown", False, False
    for line in rmd_string.split("\n"):
        if in_block and (line.strip() == "```" or re.match(END_REGEX, line)):
            in_block = False

            # collect cell_lines into a new cell
            cell = create_cell(cell_type, "\n".join(cell_lines + [line]))
            cells.append(cell)
            cell_type, cell_lines = "markdown", []

        elif line.startswith("```") or re.match(BEGIN_REGEX, line):
            in_block = True

            # collect cell_lines into a new cell
            if cell_lines:
                cell = create_cell(cell_type, "\n".join(cell_lines))
                cells.append(cell)
            cell_type = "code" if line.startswith("```{r") and "}" in line else "markdown"
            cell_lines = [line]

        else:
            cell_lines.append(line)

    # collect remaining cell lines into a new cell
    if cell_lines:
        cell = create_cell(cell_type, "\n".join(cell_lines))
        cells.append(cell)

    return cells


def collapse_empty_cells(cells):
    """
    Collapses all runs of cells with empty sources into a single cell with an empty source

    Args:
        cells (``list`` of ``Cell``): the cells

    Returns:
        ``list`` of ``Cell``: cells with runs collapsed
    """
    in_run, run_start = False, 0
    replacements = []
    for i, cell in enumerate(cells):
        if in_run and cell["source"].strip():
            if (run_start > 0 and cells[run_start-1]["source"].endswith("\n")) or cell["source"].startswith("\n"):
                replacement = []
            else:
                replacement = [create_cell("markdown", "")]
            replacements.append((run_start, i, replacement))
            in_run = False
        elif not in_run and not cell["source"].strip():
            in_run = True
            run_start = i
    
    replacements.reverse()
    for rs, re, rep in replacements:
        cells[rs:re] = rep



import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()

# get version
env = {}
with open("otter/version.py") as f:
	exec(f.read(), env)
version = env["__version__"]

# get requirements
with open("requirements.txt") as f:
	install_requires = f.readlines()

setuptools.setup(
	name = "otter-grader",
	version = version,
	author = "Christopher Pyles",
	author_email = "otter-grader@berkeley.edu",
	description = "Python and Jupyter Notebook autograder",
	long_description = long_description,
	long_description_content_type = "text/markdown",
	url = "https://github.com/ucbds-infra/otter-grader",
	license = "BSD-3-Clause",
	packages = setuptools.find_packages(exclude=["test"]),
	classifiers = [
		"Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
	],
	install_requires=install_requires,
	# scripts=["bin/otter"],
	entry_points = {
		"console_scripts": [
			"otter=otter.cli:cli", 
			"gmail_oauth2=otter.plugins.builtin.gmail_notifications.bin.gmail_oauth2:main",
		],
	},
	package_data={
		"otter.export.exporters": ["templates/*", "templates/*/*"],
		"otter.generate": ["templates/*", "templates/*/*"],
		"otter.grade": ["Dockerfile"],
	},
)



"""
Support for OK-formatted test files
"""

import os
import io
import doctest
import warnings
import pathlib

from contextlib import redirect_stderr, redirect_stdout
from textwrap import dedent

from .abstract_test import TestFile, TestCase, TestCaseResult
from ..utils import hide_outputs


def run_doctest(name, doctest_string, global_environment):
    """
    Run a single test with given ``global_environment``. Returns ``(True, '')`` if the doctest passes. 
    Returns ``(False, failure_message)`` if the doctest fails.

    Args:
        name (``str``): name of doctest
        doctest_string (``str``): doctest in string form
        global_environment (``dict``): global environment resulting from the execution of a python 
            script/notebook
    
    Returns:
        ``tuple`` of (``bool``, ``str``): results from running the test
    """
    examples = doctest.DocTestParser().parse(
        doctest_string,
        name
    )
    test = doctest.DocTest(
        [e for e in examples if isinstance(e, doctest.Example)],
        global_environment,
        name,
        None,
        None,
        doctest_string
    )

    doctestrunner = doctest.DocTestRunner(verbose=True)

    runresults = io.StringIO()
    with redirect_stdout(runresults), redirect_stderr(runresults), hide_outputs():
        doctestrunner.run(test, clear_globs=False)
    with open(os.devnull, 'w') as f, redirect_stderr(f), redirect_stdout(f):
        result = doctestrunner.summarize(verbose=True)
    # An individual test can only pass or fail
    if result.failed == 0:
        return (True, '')
    else:
        return False, runresults.getvalue()


class OKTestFile(TestFile):
    """
    A single OK-formatted test file for Otter.

    Args:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        value (``int``, optional): the point value of this test, defaults to 1
        all_or_nothing (``bool``, optional): whether the test should be graded all-or-nothing across
            cases

    Attributes:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        value (``int``): the point value of this test, defaults to 1
        all_or_nothing (``bool``): whether the test should be graded all-or-nothing across
            cases
        passed_all (``bool``): whether all of the test cases were passed
        test_case_results (``list`` of ``TestCaseResult``): a list of results for the test cases in
            ``test_cases``
        grade (``float``): the percentage of ``points`` earned for this test file as a decimal
    """

    def run(self, global_environment):
        """
        Runs tests on a given ``global_environment``

        Arguments:
            ``global_environment`` (``dict``): result of executing a Python notebook/script

        Returns:
            ``tuple`` of (``bool``, ``float`` ``otter.ok_parser.OKTest``): whether the test passed,
                the percentage score on this test, and a pointer to the current ``otter.ok_parser.OKTest`` object
        """
        for i, test_case in enumerate(self.test_cases):
            passed, result = run_doctest(self.name + ' ' + str(i), test_case.body, global_environment)
            if passed:
                result = 'Test case passed!'

            self.test_case_results.append(TestCaseResult(
                test_case = test_case,
                message = result,
                passed = passed,
            ))

    @classmethod
    def from_spec(cls, test_spec, path=""):
        # Make sure there is a name
        assert 'name' in test_spec

        # Do not support multiple suites in the same file
        assert len(test_spec['suites']) == 1

        test_suite = test_spec['suites'][0]

        # Only support doctest. I am unsure if other tests are implemented
        assert test_suite.get('type', 'doctest') == 'doctest'

        # Not setup and teardown supported
        assert not bool(test_suite.get('setup'))
        assert not bool(test_suite.get('teardown'))

        test_cases = []
        for i, test_case in enumerate(test_spec['suites'][0]['cases']):
            test_cases.append(TestCase(
                name = test_case.get('name', f"{test_spec['name']} - {i + 1}"),
                body = dedent(test_case['code']), 
                hidden = test_case.get('hidden', True),
                points = test_case.get('points', None),
                success_message = test_case.get('success_message', None),
                failure_message = test_case.get('failure_message', None)
            ))

        # resolve point values for each test case
        spec_pts = test_spec.get('points', None)
        test_cases = cls.resolve_test_file_points(spec_pts, test_cases)

        # convert path into PurePosixPath for test name
        path = str(pathlib.Path(path).as_posix())

        # grab whether the tests are all-or-nothing
        all_or_nothing = test_spec.get('all_or_nothing', True)

        return cls(test_spec['name'], path, test_cases, all_or_nothing)

    @classmethod
    def from_file(cls, path):
        """
        Parse an ok test file & return an ``OKTest``

        Args:
            path (``str``): path to ok test file

        Returns:
            ``otter.ok_parser.OKTest``: new ``OKTest`` object created from the given file
        """
        # ok test files are python files, with a global 'test' defined
        test_globals = {}
        with open(path) as f:
            exec(f.read(), test_globals)

        test_spec = test_globals['test']

        return cls.from_spec(test_spec, path=path)        



"""IPython notebook API for Otter Check"""

import datetime as dt
import inspect
import json
import os
import pickle
import re
import time
import warnings
import zipfile

from getpass import getpass
from glob import glob
from IPython import get_ipython
from IPython.display import display, HTML, Javascript
from textwrap import indent
from urllib.parse import urljoin

from .logs import LogEntry, EventType, Log
from .utils import colab_incompatible, grade_zip_file, logs_event, running_on_colab, save_notebook

from ..execute import check
from ..export import export_notebook
from ..plugins import PluginCollection
from ..test_files import GradingResults
from ..test_files.metadata_test import NOTEBOOK_METADATA_KEY


_OTTER_LOG_FILENAME = ".OTTER_LOG"
_SHELVE = False
_ZIP_NAME_FILENAME = "__zip_filename__"


class Notebook:
    """
    Notebook class for in-notebook autograding

    Args:
        nb_path(``str``, optional): path to the notebook being run
        tests_dir (``str``, optional): path to tests directory
        colab (``bool``, optional): whether this notebook is being run on Google Colab; if ``None``,
            this information is automatically parsed from IPython on creation
    """

    # overrides tests_dir arg in __init__, used for changing tests dir during grading
    _tests_dir_override = None

    @logs_event(EventType.INIT)
    def __init__(self, nb_path=None, tests_dir="./tests", colab=None):
        global _SHELVE

        if colab is None:
            colab = running_on_colab()

        if colab and not os.path.isdir(tests_dir):
            raise ValueError(f"Tests directory {tests_dir} does not exist")

        if type(self)._tests_dir_override is not None:
            self._path = type(self)._tests_dir_override
        else:
            self._path = tests_dir

        self._colab = colab
        self._notebook = nb_path
        self._addl_files = []
        self._plugin_collections = {}

        # assume using otter service if there is a .otter file
        otter_configs = glob("*.otter")
        if otter_configs:
            # check that there is only 1 config
            assert len(otter_configs) == 1, "More than 1 otter config file found"

            # load in config file
            with open(otter_configs[0], encoding="utf-8") as f:
                self._config = json.load(f)

            _SHELVE = self._config.get("save_environment", False)
            self._ignore_modules = self._config.get("ignore_modules", [])
            self._vars_to_store = self._config.get("variables", None)

            self._notebook = self._config["notebook"]

    def _log_event(self, event_type, results=[], question=None, success=True, error=None, shelve_env={}):
        """
        Logs an event

        Args:
            event_type (``otter.logs.EventType``): the type of event
            results (``list[otter.test_files.abstract_test.TestFile]``, optional):
                the results of any checks recorded by the entry
            question (``str``, optional): the question name for this check
            success (``bool``, optional): whether the operation was successful
            error (``Exception``, optional): the exception thrown by the operation, if applicable
        """
        entry = LogEntry(
            event_type,
            results=results,
            question=question,
            success=success,
            error=error
        )

        if _SHELVE and event_type == EventType.CHECK:
            entry.shelve(
                shelve_env,
                delete=True,
                filename=_OTTER_LOG_FILENAME,
                ignore_modules=self._ignore_modules,
                variables=self._vars_to_store
            )

        entry.flush_to_file(_OTTER_LOG_FILENAME)

    def _resolve_nb_path(self, nb_path):
        """
        Attempts to resolve the path to the notebook being run. If ``nb_path`` is ``None``, ``self._notebook``
        is checked, then the working directory is searched for ``.ipynb`` files. If none are found, or 
        more than one is found, a ``ValueError`` is raised.

        Args:
            nb_path (``Optional[str]``): path to the notebook
        
        Returns:
            ``str``: resolved notebook path
        
        Raises:
            ``ValueError``: if no notebooks or too many notebooks are found.
        """
        if nb_path is None and self._notebook is not None:
            nb_path = self._notebook
            if not os.path.isfile(nb_path):
                raise ValueError(f"Expected a notebook file named '{nb_path}' but no such file found")

        elif nb_path is None and glob("*.ipynb"):
            notebooks = glob("*.ipynb")
            assert len(notebooks) == 1, "nb_path not specified and > 1 notebook in working directory"
            nb_path = notebooks[0]

        elif nb_path is None:
            raise ValueError("Could not resolve notebook path")

        return nb_path

    @logs_event(EventType.CHECK)
    def check(self, question, global_env=None):
        """
        Runs tests for a specific question against a global environment. If no global environment
        is provided, the test is run against the calling frame's environment.

        Args:
            question (``str``): name of question being graded
            global_env (``dict``, optional): global environment resulting from execution of a single
                notebook

        Returns:
            ``otter.test_files.abstract_test.TestFile``: the grade for the question
        """
        if os.path.isdir(self._path) and os.path.isfile(os.path.join(self._path, question + ".py")):
            test_path = os.path.join(self._path, question + ".py")
            test_name = None

        elif self._colab:
            raise ValueError(f"Test {question} does not exist")

        else:
            test_path = self._resolve_nb_path(None)
            test_name = question

        # ensure that desired test exists
        assert os.path.isfile(test_path), "Test {} does not exist".format(question)

        # pass the correct global environment
        if global_env is None:
            global_env = inspect.currentframe().f_back.f_back.f_globals

        # run the check
        result = check(test_path, test_name, global_env)

        return question, result, global_env

    @colab_incompatible
    def run_plugin(self, plugin_name, *args, nb_path=None, **kwargs):
        """
        Runs the plugin ``plugin_name`` with the specified arguments. Use ``nb_path`` if the path
        to the notebook is not configured.

        Args:
            plugin_name (``str``): importable name of an Otter plugin that implements the 
                ``from_notebook`` hook
            *args: arguments to be passed to the plugin
            nb_path (``str``, optional): path to the notebook
            **kwargs: keyword arguments to be passed to the plugin

        """
        nb_path = self._resolve_nb_path(nb_path)
        if plugin_name in self._plugin_collections:
            pc = self._plugin_collections[plugin_name]
        else:
            pc = PluginCollection([plugin_name], nb_path, {})
            self._plugin_collections[plugin_name] = pc
        pc.run("from_notebook", *args, **kwargs)

    @colab_incompatible
    @logs_event(EventType.TO_PDF)
    def to_pdf(self, nb_path=None, filtering=True, pagebreaks=True, display_link=True, force_save=False):
        """
        Exports a notebook to a PDF using Otter Export

        Args:
            nb_path (``str``, optional): path to the notebook we want to export; will attempt to infer
                if not provided
            filtering (``bool``, optional): set true if only exporting a subset of notebook cells to PDF
            pagebreaks (``bool``, optional): if true, pagebreaks are included between questions
            display_link (``bool``, optional): whether or not to display a download link
            force_save (``bool``, optional): whether or not to display JavaScript that force-saves the
                notebook (only works in Jupyter Notebook classic, not JupyterLab)
        """
        nb_path = self._resolve_nb_path(nb_path)

        if force_save:
            saved = save_notebook(nb_path)
            if not saved:
                warnings.warn(
                    "Could not force-save notebook; the results of this call will be based on the last "
                    "saved version of this notebook."
                )

        # convert(nb_path, filtering=filtering, filter_type=filter_type)
        export_notebook(nb_path, filtering=filtering, pagebreaks=pagebreaks)

        if display_link:
            # create and display output HTML
            out_html = """
            <p>Your file has been exported. Download it by right-clicking
            <a href="{}" target="_blank">here</a> and selecting <strong>Save Link As</strong>.
            """.format(nb_path[:-5] + "pdf")

            display(HTML(out_html))
        self._log_event(EventType.TO_PDF)

    @colab_incompatible
    def add_plugin_files(self, plugin_name, *args, nb_path=None, **kwargs):
        """
        Runs the ``notebook_export`` event of the plugin ``plugin_name`` and tracks the file paths
        it returns to be included when calling ``Notebook.export``.

        Args:
            plugin_name (``str``): importable name of an Otter plugin that implements the 
                ``from_notebook`` hook
            *args: arguments to be passed to the plugin
            nb_path (``str``, optional): path to the notebook
            **kwargs: keyword arguments to be passed to the plugin        
        """
        nb_path = self._resolve_nb_path(nb_path)
        if plugin_name in self._plugin_collections:
            pc = self._plugin_collections[plugin_name]
        else:
            pc = PluginCollection([plugin_name], nb_path, {})
            self._plugin_collections[plugin_name] = pc
        addl_files = pc.run("notebook_export", *args, **kwargs)[0]
        if addl_files is None:
            return
        self._addl_files.extend(addl_files)

    @colab_incompatible
    @logs_event(EventType.END_EXPORT)
    def export(self, nb_path=None, export_path=None, pdf=True, filtering=True, pagebreaks=True, files=[], 
            display_link=True, force_save=False, run_tests=False):
        """
        Exports a submission to a zip file. Creates a submission zipfile from a notebook at ``nb_path``,
        optionally including a PDF export of the notebook and any files in ``files``.

        Args:
            nb_path (``str``, optional): path to the notebook we want to export; will attempt to infer
                if not provided
            export_path (``str``, optional): path at which to write zipfile; defaults to notebook's
                name + ``.zip``
            pdf (``bool``, optional): whether a PDF should be included
            filtering (``bool``, optional): whether the PDF should be filtered; ignored if ``pdf`` is
                ``False``
            pagebreaks (``bool``, optional): whether pagebreaks should be included between questions
                in the PDF
            files (``list`` of ``str``, optional): paths to other files to include in the zip file
            display_link (``bool``, optional): whether or not to display a download link
            force_save (``bool``, optional): whether or not to display JavaScript that force-saves the
                notebook (only works in Jupyter Notebook classic, not JupyterLab)
        """
        self._log_event(EventType.BEGIN_EXPORT)

        nb_path = self._resolve_nb_path(nb_path)

        if force_save:
            saved = save_notebook(nb_path)
            if not saved:
                warnings.warn(
                    "Could not force-save notebook; the results of this call will be based on the last "
                    "saved version of this notebook."
                )

        try:
            with open(nb_path) as f:
                assert len(f.read().strip()) > 0, \
                    f"Notebook {nb_path} is empty. Please save and checkpoint your notebook and rerun this cell."

        except UnicodeDecodeError:
            with open(nb_path, "r", encoding="utf-8") as f:
                assert len(f.read().strip()) > 0, \
                    f"Notebook {nb_path} is empty. Please save and checkpoint your notebook and rerun this cell."

        timestamp = dt.datetime.now().strftime("%Y_%m_%dT%H_%M_%S_%f")
        if export_path is None:
            zip_path = ".".join(nb_path.split(".")[:-1]) + "_" + timestamp + ".zip"
        else:
            zip_path = export_path

        zf = zipfile.ZipFile(zip_path, mode="w")
        zf.write(nb_path)

        if pdf:
            pdf_path = ".".join(nb_path.split(".")[:-1]) + ".pdf"
            # convert(nb_path, filtering=filtering, filter_type=filter_type)
            export_notebook(nb_path, filtering=filtering, pagebreaks=pagebreaks)
            if os.path.isfile(pdf_path):
                zf.write(pdf_path)
            else:
                warnings.warn("Could not locate a PDF to include")

        if os.path.isfile(_OTTER_LOG_FILENAME):
            zf.write(_OTTER_LOG_FILENAME)

        zf.writestr(_ZIP_NAME_FILENAME, os.path.basename(zip_path))

        if glob("*.otter"):
            assert len(glob("*.otter")) == 1, "Too many .otter files (max 1 allowed)"
            zf.write(glob("*.otter")[0])

        for file in files:
            zf.write(file)
        
        for file in self._addl_files:
            zf.write(file)

        zf.close()

        if run_tests:
            print("Running your submission against local test cases...\n")
            results = grade_zip_file(zip_path, nb_path, self._path)
            print(
                "Your submission received the following results when run against " + \
                "available test cases:\n\n" + indent(results.summary(), "    ")
            )

        if display_link:
            # create and display output HTML
            out_html = """
            <p>Your submission has been exported. Click <a href="{}" download="{}" target="_blank">here</a>
            to download the zip file.</p>
            """.format(zip_path, zip_path)

            display(HTML(out_html))

    @logs_event(EventType.END_CHECK_ALL)
    def check_all(self):
        """
        Runs all tests on this notebook. Tests are run against the current global environment, so any
        tests with variable name collisions will fail.
        """
        # TODO: this should use functions in execute.py to run tests in-sequence so that variable
        # name collisions are accounted for
        self._log_event(EventType.BEGIN_CHECK_ALL)

        # TODO: this is a janky way of resolving where the tests are. Formalize a method of 
        # determining this and put it into a method in e.g. utils.py
        tests = [os.path.split(file)[1][:-3] for file in glob(os.path.join(self._path, "*.py")) \
            if "__init__.py" not in file]
        if len(tests) == 0:
            nb_path = self._resolve_nb_path(None)
            with open(nb_path, encoding="utf-8") as f:
                nb = json.load(f)
            tests = list(nb["metadata"][NOTEBOOK_METADATA_KEY]["tests"].keys())

        global_env = inspect.currentframe().f_back.f_back.f_globals
        results = []
        if not _SHELVE:
            for test_name in sorted(tests):
                results.append(self.check(test_name, global_env))
        else:
            log = Log.from_file(_OTTER_LOG_FILENAME, ascending=False)
            for file in sorted(tests):
                if "__init__.py" not in file:
                    test_name = os.path.splitext(os.path.split(file)[1])[0]

                    entry = log.get_question_entry(test_name)
                    env = entry.unshelve()
                    global_env.update(env)
                    del locals()["env"]

                    result = self.check(test_name, global_env)
                    results.append((test_name, result))

        return GradingResults(results)



"""Default configurations for autograding"""

from ...utils import convert_config_description_dict

# a dictionary for documenting each configuration and its default value; the dictionary of actual
# configurations that gets imported is generated from this dictionary
DEFAULT_OPTIONS_WITH_DESCRIPTIONS = [
    {
        "key": "score_threshold",
        "description": "a score threshold for pass-fail assignments",
        "default": None,
    },
    {
        "key": "points_possible",
        "description": "a custom total score for the assignment; if unspecified the sum of question point values is used.",
        "default": None,
    },
    {
        "key": "show_stdout",
        "description": "whether to display the autograding process stdout to students on Gradescope",
        "default": False,
    },
    {
        "key": "show_hidden",
        "description": "whether to display the results of hidden tests to students on Gradescope",
        "default": False,
    },
    {
        "key": "show_all_public",
        "description": "whether to display all test results if all tests are public tests",
        "default": False,
    },
    {
        "key": "seed",
        "description": "a random seed for intercell seeding",
        "default": None,
    },
    {
        "key": "seed_variable",
        "description": "a variable name to override with the seed",
        "default": None,
    },
    {
        "key": "grade_from_log",
        "description": "whether to re-assemble the student's environment from the log rather than by re-executing their submission",
        "default": False,
    },
    {
        "key": "serialized_variables",
        "description": "a mapping of variable names to type strings for validating a deserialized student environment",
        "default": {},
    },
    {
        "key": "pdf",
        "description": "whether to generate a PDF of the notebook when not using Gradescope auto-upload",
        "default": False,
    },
    {
        "key": "token",
        "description": "a Gradescope token for uploading a PDF of the notebook",
        "default": None,
    },
    {
        "key": "course_id",
        "description": "a Gradescope course ID for uploading a PDF of the notebook",
        "default": "None",
    },
    {
        "key": "assignment_id",
        "description": "a Gradescope assignment ID for uploading a PDF of the notebook",
        "default": "None",
    },
    {
        "key": "filtering",
        "description": "whether the generated PDF should have cells filtered out",
        "default": False,
    },
    {
        "key": "pagebreaks",
        "description": "whether the generated PDF should have pagebreaks between filtered sectios",
        "default": False,
    },
    {
        "key": "debug",
        "description": "whether to run the autograder in debug mode (without ignoring errors)",
        "default": False,
    },
    {
        "key": "autograder_dir",
        "description": "the directory in which autograding is taking place",
        "default": "/autograder",
    },
    {
        "key": "lang",
        "description": "the language of the assignment; one of {'python', 'r'}",
        "default": "python",
    },
    {
        "key": "miniconda_path",
        "description": "the path to the miniconda install directory",
        "default": "/root/miniconda3",
    },
    {
        "key": "plugins",
        "description": "a list of plugin names and configuration details for grading",
        "default": [],
    },
    {
        "key": "logo",
        "description": "whether to print the Otter logo to stdout",
        "default": True,
    },
    {
        "key": "print_summary",
        "description": "whether to print the grading summary",
        "default": True,
    },
    {
        "key": "print_score",
        "description": "whether to print out the submission score in the grading summary",
        "default": True,
    },
    {
        "key": "zips",
        "description": "whether zip files are being graded",
        "default": False,
    },
    {
        "key": "channel_priority_strict",
        "description": "whether to set conda's channel_priority config to strict in the setup.sh file",
        "default": True,
    },
]

DEFAULT_OPTIONS = convert_config_description_dict(DEFAULT_OPTIONS_WITH_DESCRIPTIONS)



"""Master notebook transformer for Otter Assign"""

import copy
import nbformat
import os
import pathlib

from .blocks import BlockType, get_cell_config, is_assignment_config_cell, is_block_boundary_cell
from .cell_generators import (
    add_export_tag_to_cell, gen_init_cell, gen_markdown_response_cell, gen_export_cells, 
    gen_check_all_cell
)
from .questions import add_point_value_info_to_cell, create_question_config
from .r_adapter.cell_generators import gen_export_cells as gen_ottr_export_cells
from .solutions import has_seed, SOLUTION_CELL_TAG
from .tests import any_public_tests, determine_question_point_value
from .utils import add_tag, AssignNotebookFormatException, EmptyCellException, get_source, \
    is_cell_type, is_ignore_cell


def transform_notebook(nb, assignment):
    """
    Converts a master notebook to an Otter-formatted solutions notebook, parsing test cells into
    dictionaries ready to be written as OK test files.

    Args:
        nb (``nbformat.NotebookNode``): the master notebook
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``tuple[nbformat.NotebookNode, dict]``: the transformed notebook and a dictionary mapping 
        test names to their parsed contents

    """
    transformed_cells, test_files = get_transformed_cells(nb['cells'], assignment)

    if assignment.init_cell and assignment.is_python:
        transformed_cells = [gen_init_cell(assignment.master.name, assignment.colab)] + transformed_cells

    if assignment.check_all_cell and assignment.is_python:
        transformed_cells += gen_check_all_cell()

    if assignment.export_cell:
        export_cell = assignment.export_cell
        if export_cell is True:
            export_cell = {}

        # TODO: convert export_cell to default dict
        if assignment.is_python:
            transformed_cells += gen_export_cells(
                export_cell.get('instructions', ''), 
                pdf = export_cell.get('pdf', True),
                filtering = export_cell.get('filtering', True),
                force_save = export_cell.get('force_save', False),
                run_tests = export_cell.get('run_tests', False))

        elif assignment.is_r:
            transformed_cells += gen_ottr_export_cells(
                assignment.notebook_basename,
                export_cell.get('instructions', ''))

    transformed_nb = copy.deepcopy(nb)
    transformed_nb['cells'] = transformed_cells

    return transformed_nb, test_files


def get_transformed_cells(cells, assignment):
    """
    Takes in a list of cells from the master notebook and returns a list of cells for the solutions
    notebook. 

    Replaces test cells with a cell calling ``otter.Notebook.check``, inserts Markdown
    response cells for manual questions with Markdown solutions, and comments out question metadata 
    in question cells, among other things.

    Args:
        cells (``list[nbformat.NotebookNode]``): original code cells
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``tuple[list, dict]``: list of cleaned notebook cells and a dictionary mapping test names to 
        their parsed contents
    """
    if assignment.is_r:
        from otter.assign.r_adapter.tests import read_test, gen_test_cell
    else:
        from otter.assign.tests import read_test, gen_test_cell

    curr_block = []  # allow nested blocks
    transformed_cells = []
    test_files = {}

    question_metadata, test_cases, has_prompt, no_solution, last_question_md_cell = \
        {}, [], False, False, None

    solution_has_md_cells, prompt_insertion_index = False, None

    need_begin_export, need_end_export = False, False

    for i, cell in enumerate(cells):

        if is_ignore_cell(cell):
            continue

        # check for assignment config
        if is_assignment_config_cell(cell):
            assignment.update(get_cell_config(cell))
            continue

        # check for an end to the current block
        if len(curr_block) > 0 and is_block_boundary_cell(cell, curr_block[-1], end=True):
            block_type = curr_block.pop()  # remove the block

            if block_type is BlockType.QUESTION:

                if question_metadata["manual"] or question_metadata["export"]:
                    need_end_export = True

                # generate a check cell
                if test_cases:
                    check_cell = gen_test_cell(question_metadata, test_cases, test_files, assignment)

                    # only add to notebook if there's a response cell or if there are public tests
                    # don't add cell if the 'check_cell' key of quetion_metadata is false
                    if (not no_solution or any_public_tests(test_cases)) and \
                            question_metadata["check_cell"]:
                        transformed_cells.append(check_cell)

                # add points to question cell if specified
                if assignment.show_question_points and last_question_md_cell is not None:
                    points = determine_question_point_value(question_metadata, test_cases)
                    transformed_cells[last_question_md_cell] = \
                        add_point_value_info_to_cell(transformed_cells[last_question_md_cell], points)

                # TODO: reformat this state update
                question_metadata, test_cases, has_prompt, no_solution, last_question_md_cell = \
                    {}, [], False, False, None
                solution_has_md_cells, prompt_insertion_index = False, None

            elif block_type is BlockType.SOLUTION:
                if not has_prompt and solution_has_md_cells:
                    if prompt_insertion_index is None:
                        # TODO: make this error nicer?
                        raise RuntimeError("Could not find prompt insertion index")
                    transformed_cells.insert(prompt_insertion_index, gen_markdown_response_cell())
                    has_prompt = True

            continue  # if this is an end to the last nested block, we're OK

        # check for invalid block ends
        for block_type in BlockType:
            if is_block_boundary_cell(cell, block_type, end=True):

                # if a child is missing an end block cell, raise an error
                if block_type in curr_block:
                    raise AssignNotebookFormatException(
                        f"Found an end { block_type.value } cell with an un-ended child " + \
                            f"{ curr_block[-1].value } block", question_metadata, i)

                # otherwise raise an error for an end with no begin
                else:
                    raise AssignNotebookFormatException(
                        f"Found an end { block_type.value } cell with no begin block cell",
                        question_metadata, i)

        # check for begin blocks
        found_begin = False
        for block_type in BlockType:
            if is_block_boundary_cell(cell, block_type):
                found_begin = True
                break

        if found_begin:
            if len(curr_block) == 0 and block_type is not BlockType.QUESTION:
                raise AssignNotebookFormatException(
                    f"Found a begin { block_type.value } cell outside a question", 
                    question_metadata, i)
            elif len(curr_block) > 0 and block_type is BlockType.QUESTION:
                raise AssignNotebookFormatException(
                    f"Found a begin { block_type.value } cell inside another question", 
                    question_metadata, i)
            elif len(curr_block) > 1:
                raise AssignNotebookFormatException(
                    f"Found a begin { block_type.value } cell inside a { curr_block[-1].value } " + \
                        f"block", question_metadata, i)
            elif block_type is BlockType.PROMPT and has_prompt:  # has_prompt was set by the solution block
                raise AssignNotebookFormatException(
                    "Found a prompt block after a solution block", question_metadata, i)

            # if not an invalid begin cell, update state
            if block_type is BlockType.PROMPT:
                has_prompt = True

            elif block_type is BlockType.SOLUTION and not has_prompt:
                prompt_insertion_index = len(transformed_cells)

            elif block_type is BlockType.TESTS and not has_prompt:
                no_solution = True

            elif block_type is BlockType.QUESTION:
                question_metadata = get_cell_config(cell)
                if not isinstance(question_metadata, dict):
                    raise AssignNotebookFormatException(
                        "Found a begin question cell with no config", question_metadata, i)
                question_metadata = create_question_config(question_metadata)
                if question_metadata["manual"] or question_metadata["export"]:
                    need_begin_export = True

            curr_block.append(block_type)
            continue

        # if in a block, process the current cell
        if len(curr_block) > 0:
            if curr_block[-1] == BlockType.TESTS:
                if not is_cell_type(cell, "code"):
                    raise AssignNotebookFormatException(
                        "Found a non-code cell in tests block", question_metadata, i)
                test_case = read_test(cell, question_metadata, assignment)
                test_cases.append(test_case)
                continue

            elif curr_block[-1] == BlockType.SOLUTION:
                cell = add_tag(cell, SOLUTION_CELL_TAG)

                if is_cell_type(cell, "markdown"):
                    solution_has_md_cells = True

            elif curr_block[-1] == BlockType.QUESTION and is_cell_type(cell, "markdown"):
                last_question_md_cell = len(transformed_cells)

        # add export tags if needed
        export_delim_cell = None
        if need_begin_export:
            if is_cell_type(cell, "markdown"):
                cell = add_export_tag_to_cell(cell)
            else:
                export_delim_cell = nbformat.v4.new_markdown_cell()
                export_delim_cell = add_export_tag_to_cell(export_delim_cell)
            need_begin_export = False
        if need_end_export:
            if is_cell_type(cell, "markdown"):
                cell = add_export_tag_to_cell(cell, end=True)
            else:
                if export_delim_cell is None:
                    export_delim_cell = nbformat.v4.new_markdown_cell()
                export_delim_cell = add_export_tag_to_cell(export_delim_cell, end=True)
            need_end_export = False

        if export_delim_cell is not None:
            transformed_cells.append(export_delim_cell)

        if has_seed(cell):
            assignment.seed_required = True

        # this is just a normal cell so add it to transformed_cells
        transformed_cells.append(cell)

    # if the last cell was the end of a manually-graded question, add a close export tag
    if need_end_export:
        transformed_cells.append(add_export_tag_to_cell(nbformat.v4.new_markdown_cell(), end=True))

    return transformed_cells, test_files
